"""
Contains modules that implement the core classes of the project. To make these
classes easily available for users, they are all already imported into the
namespace `iracema`. For instance, the class ``Audio`` is available in the
module ``iracema.core.audio``, but it can be accessed using the shorter name
``iracema.Audio``.
"""
